@@ -17,7 +17,7 @@ countpattern <- function(x, matching=FALSE)
          xpat <- 2*xpat+x[,i]
      xpat <- xpat+1
 
-     pat <- tabulate(xpat, nb=2^nvar)
+     pat <- tabulate(xpat, nbins=2^nvar)
      names(pat) <- namespat
 
      if (matching)
