@@ -605,7 +605,9 @@ function(x, data, formula = NULL, fill = TRUE,
             names(lis)[1:2] <- colnames(sub)
             new <- expand.grid(lis)[, labels(terms(x))]
             preds <- predict(x, new)
-            filled.contour(xr, yr, matrix(as.numeric(preds), nr = length(xr), byrow = TRUE),
+            filled.contour(xr, yr,
+                           matrix(as.numeric(preds),
+                                  nrow = length(xr), byrow = TRUE),
                            plot.axes = {
                                axis(1)
                                axis(2)
